use cat_box::{draw_text, get_keyboard_state, Game, Sprite, SpriteCollection};
use rand::thread_rng;
use rand::Rng;
use sdl2::keyboard::Scancode;
use std::time::Instant;

#[derive(Copy, Clone, Debug, Eq, PartialEq)]
enum Direction {
    Up,
    Down,
    Left,
    Right,
}

macro_rules! set_if_not_opp {
    ($i:ident, $e:expr, $opp:expr, $v:expr) => {
        if $i != $opp {
            $v.push($e);
            $i = $e;
        }
    };
}

fn main() {
    //hardcoding size to make this game as inconvient as you can
    let game = Game::new("Snake", 1036, 1036);

    //NO NEED FOR LOCAL RUNNING, ALL IN EXECUTABLE
    let snake_bytes = include_bytes!("../snakecell.png");
    let apple_bytes = include_bytes!("../apple.png");
    let wall_bytes = include_bytes!("../wall.png");
    let grid_bytes = include_bytes!("../grid2.png");

    //INPROG GRID SYSTEM INSTEAD OF DIRECT POSITION MANAGEMENT

    //SNAKE TRAIN
    let snake_boxes: Vec<(i32, i32)> = vec![(13, 13), (14, 13)];
    let mut snake = SpriteCollection::with_capacity(snake_boxes.len());
    for (x, y) in snake_boxes {
        let s = Sprite::from_bytes(snake_bytes, x * 37, y * 37).unwrap();
        snake.push(s);
    }

    //RANDOM APPLE SPAWNING
    let mut apple = {
        let x = thread_rng().gen_range(2..=26);
        let y = thread_rng().gen_range(2..=26);

        Sprite::from_bytes(apple_bytes, x * 37, y * 37).unwrap()
    };

    //I am lazy don't @ me and it WORKS
    let mut walls = {
        let mut wall_boxes = SpriteCollection::with_capacity(27);
        let mut num = 999;
        while num >= 1 {
            let wall = Sprite::from_bytes(wall_bytes, num, 37).unwrap();
            wall_boxes.push(wall);
            num -= 37;
        }
        while num < 999 {
            num += 37;
            let wall = Sprite::from_bytes(wall_bytes, num, 999).unwrap();
            wall_boxes.push(wall);
        }
        while num >= 1 {
            let wall = Sprite::from_bytes(wall_bytes, 37, num).unwrap();
            wall_boxes.push(wall);
            num -= 37;
        }
        while num < 999 {
            num += 37;
            let wall = Sprite::from_bytes(wall_bytes, 999, num).unwrap();
            wall_boxes.push(wall);
        }
        wall_boxes
    };

    //GRID IMAGE
    let mut grid = Sprite::from_bytes(grid_bytes, 499 + (37 / 2), 499 + (37 / 2)).unwrap();

    //DEFAULT DIRECTION
    let mut dir = Direction::Left;
    let mut dir2 = Direction::Left;
    let mut queue: Vec<Direction> = Vec::new();
    let mut score = 0u64;
    let mut now = Instant::now();

    game.run(|ctx| {
        let keys = get_keyboard_state(ctx).keys;

        for key in keys {
            use Direction::*;
            match key {
                Scancode::Q | Scancode::Escape => {
                    println!("Game over!");
                    println!("Your score was: {}", score);
                    game.terminate();
                }
                Scancode::W | Scancode::Up => set_if_not_opp!(dir, Up, Down, queue),
                Scancode::A | Scancode::Left => set_if_not_opp!(dir, Left, Right, queue),
                Scancode::S | Scancode::Down => set_if_not_opp!(dir, Down, Up, queue),
                Scancode::D | Scancode::Right => set_if_not_opp!(dir, Right, Left, queue),
                _ => (),
            };
        }

        if now.elapsed().as_millis() >= 125 {
            dir2 = dir.clone();
            {
                let mut last_part = snake[0].position();

                for s in snake.iter().skip(1) {
                    let (lastx, lasty) = last_part;
                    let (x, y) = s.position();
                    let (xdiff, ydiff) = (lastx - x, y - lasty);
                    last_part = s.position();
                    s.translate((xdiff, ydiff));
                }
            }
            // The snake head needs to be moved after the body or else the body will just collapse into the head

            match dir2 {
                Direction::Up => {
                    snake[0].translate((0, 37));
                }
                Direction::Left => {
                    snake[0].translate((-37, 0));
                }
                Direction::Down => {
                    snake[0].translate((0, -37));
                }
                Direction::Right => {
                    snake[0].translate((37, 0));
                }
            };
            //queue.remove(0);

            {
                let hitted_self =
                    cat_box::physics::check_for_collision_with_collection(&snake[0], &snake);
                let hitted_wall =
                    cat_box::physics::check_for_collision_with_collection(&snake[0], &walls);

                if hitted_self.len() > 1 || hitted_wall.len() > 0 {
                    println!("Game over!");
                    println!("Your score was: {}", score);
                    game.terminate();
                }
            }

            if !cat_box::physics::check_for_collision_with_collection(&apple, &snake).is_empty() {
                let x = thread_rng().gen_range(2..=26) * 37;
                let y = thread_rng().gen_range(2..=26) * 37;

                let (currx, curry) = apple.position();
                let (xdiff, ydiff) = (x - currx, curry - y);
                apple.translate((xdiff, ydiff));
                let second_to_last = snake[snake.len() - 2].position();
                let last = snake[snake.len() - 1].position();

                let direc = check_direction(last, second_to_last);

                let (newx, newy) = match direc {
                    Direction::Left => (last.0 - 37, last.1),
                    Direction::Right => (last.0 + 37, last.1),
                    Direction::Up => (last.0, last.1 - 37),
                    Direction::Down => (last.0, last.1 + 37),
                };

                let s = Sprite::new("snakecell.png", newx, newy).unwrap();
                snake.push(s);

                score += 1;
            }

            now = Instant::now();
        }

        // So that the snake doesn't move at super speed
        //std::thread::sleep(Duration::from_millis(125));

        snake.draw(ctx).unwrap();
        walls.draw(ctx).unwrap();
        grid.draw(ctx).unwrap();
        apple.draw(ctx).unwrap();

        //draw score above all else
        draw_text(
            ctx,
            format!("Score: {} Dir: {:?}", score, dir),
            "ibm_bios-2y.ttf",
            36,
            (100, 100),
            cat_box::TextMode::Transparent {
                colour: (213, 123, 212),
            },
        )
        .unwrap();
    })
    .unwrap();
}

fn check_direction(square1: (i32, i32), square2: (i32, i32)) -> Direction {
    if square1.0 < square2.0 {
        return Direction::Left;
    } else if square1.0 > square2.0 {
        return Direction::Right;
    };
    if square1.1 > square2.1 {
        return Direction::Up;
    } else if square1.1 < square2.1 {
        return Direction::Down;
    };

    unreachable!(
        "why are we still here? just to suffer? (this should never ever ever happen, by the way)"
    );
}
